# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Keras initializer serialization / deserialization.

"""

from __future__ import print_function as _print_function

from tensorflow.python.keras.initializers import ConstantV2 as Constant
from tensorflow.python.keras.initializers import ConstantV2 as constant
from tensorflow.python.keras.initializers import GlorotNormalV2 as GlorotNormal
from tensorflow.python.keras.initializers import GlorotNormalV2 as glorot_normal
from tensorflow.python.keras.initializers import GlorotUniformV2 as GlorotUniform
from tensorflow.python.keras.initializers import GlorotUniformV2 as glorot_uniform
from tensorflow.python.keras.initializers import IdentityV2 as Identity
from tensorflow.python.keras.initializers import IdentityV2 as identity
from tensorflow.python.keras.initializers import InitializerV2 as Initializer
from tensorflow.python.keras.initializers import OnesV2 as Ones
from tensorflow.python.keras.initializers import OnesV2 as ones
from tensorflow.python.keras.initializers import OrthogonalV2 as Orthogonal
from tensorflow.python.keras.initializers import OrthogonalV2 as orthogonal
from tensorflow.python.keras.initializers import RandomNormalV2 as RandomNormal
from tensorflow.python.keras.initializers import RandomUniformV2 as RandomUniform
from tensorflow.python.keras.initializers import TruncatedNormalV2 as TruncatedNormal
from tensorflow.python.keras.initializers import VarianceScalingV2 as VarianceScaling
from tensorflow.python.keras.initializers import ZerosV2 as Zeros
from tensorflow.python.keras.initializers import ZerosV2 as zeros
from tensorflow.python.keras.initializers import deserialize
from tensorflow.python.keras.initializers import get
from tensorflow.python.keras.initializers import he_normalV2 as he_normal
from tensorflow.python.keras.initializers import he_uniformV2 as he_uniform
from tensorflow.python.keras.initializers import lecun_normalV2 as lecun_normal
from tensorflow.python.keras.initializers import lecun_uniformV2 as lecun_uniform
from tensorflow.python.keras.initializers import serialize

del _print_function
