# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Evaluation-related metrics.
"""

from __future__ import print_function as _print_function

from tensorflow.python.ops.metrics import accuracy
from tensorflow.python.ops.metrics import auc
from tensorflow.python.ops.metrics import average_precision_at_k
from tensorflow.python.ops.metrics import false_negatives
from tensorflow.python.ops.metrics import false_negatives_at_thresholds
from tensorflow.python.ops.metrics import false_positives
from tensorflow.python.ops.metrics import false_positives_at_thresholds
from tensorflow.python.ops.metrics import mean
from tensorflow.python.ops.metrics import mean_absolute_error
from tensorflow.python.ops.metrics import mean_cosine_distance
from tensorflow.python.ops.metrics import mean_iou
from tensorflow.python.ops.metrics import mean_per_class_accuracy
from tensorflow.python.ops.metrics import mean_relative_error
from tensorflow.python.ops.metrics import mean_squared_error
from tensorflow.python.ops.metrics import mean_tensor
from tensorflow.python.ops.metrics import percentage_below
from tensorflow.python.ops.metrics import precision
from tensorflow.python.ops.metrics import precision_at_k
from tensorflow.python.ops.metrics import precision_at_thresholds
from tensorflow.python.ops.metrics import precision_at_top_k
from tensorflow.python.ops.metrics import recall
from tensorflow.python.ops.metrics import recall_at_k
from tensorflow.python.ops.metrics import recall_at_thresholds
from tensorflow.python.ops.metrics import recall_at_top_k
from tensorflow.python.ops.metrics import root_mean_squared_error
from tensorflow.python.ops.metrics import sensitivity_at_specificity
from tensorflow.python.ops.metrics import sparse_average_precision_at_k
from tensorflow.python.ops.metrics import sparse_precision_at_k
from tensorflow.python.ops.metrics import specificity_at_sensitivity
from tensorflow.python.ops.metrics import true_negatives
from tensorflow.python.ops.metrics import true_negatives_at_thresholds
from tensorflow.python.ops.metrics import true_positives
from tensorflow.python.ops.metrics import true_positives_at_thresholds

del _print_function

import sys as _sys
from tensorflow.python.util import deprecation_wrapper as _deprecation_wrapper

if not isinstance(_sys.modules[__name__], _deprecation_wrapper.DeprecationWrapper):
  _sys.modules[__name__] = _deprecation_wrapper.DeprecationWrapper(
      _sys.modules[__name__], "metrics")
