# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.experimental namespace.
"""

from __future__ import print_function as _print_function

from tensorflow.python.feature_column.feature_column_lib import SequenceFeatures
from tensorflow.python.keras.layers import PeepholeLSTMCell
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import CosineDecay
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import CosineDecayRestarts
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import LinearCosineDecay
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import NoisyLinearCosineDecay
from tensorflow.python.keras.saving import export_saved_model
from tensorflow.python.keras.saving import load_from_saved_model
from tensorflow.python.keras.utils.data_utils import terminate_keras_multiprocessing_pools

del _print_function

import sys as _sys
from tensorflow.python.util import deprecation_wrapper as _deprecation_wrapper

if not isinstance(_sys.modules[__name__], _deprecation_wrapper.DeprecationWrapper):
  _sys.modules[__name__] = _deprecation_wrapper.DeprecationWrapper(
      _sys.modules[__name__], "keras.experimental")
