# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.lookup namespace.
"""

from __future__ import print_function as _print_function

from tensorflow._api.v1.compat.v1.lookup import experimental
from tensorflow.python.ops.lookup_ops import KeyValueTensorInitializer
from tensorflow.python.ops.lookup_ops import StaticHashTableV1 as StaticHashTable
from tensorflow.python.ops.lookup_ops import StaticVocabularyTableV1 as StaticVocabularyTable
from tensorflow.python.ops.lookup_ops import TextFileIndex
from tensorflow.python.ops.lookup_ops import TextFileInitializer

del _print_function
