# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.optimizers.schedules namespace.
"""

from __future__ import print_function as _print_function

from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import ExponentialDecay
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import InverseTimeDecay
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import LearningRateSchedule
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import PiecewiseConstantDecay
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import PolynomialDecay
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import deserialize
from tensorflow.python.keras.optimizer_v2.learning_rate_schedule import serialize

del _print_function
